/* 
 * jsgb.CPU.js v0.02 - GB CPU Emulator for JSGB, a JavaScript GameBoy Emulator
 * Copyright (C) 2009 Pedro Ladaria <Sonic1980 at Gmail dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The full license is available at http://www.gnu.org/licenses/gpl.html
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * ____________________________________________________________________________
 *
 * About this emulator:
 *
 * The nintendo GameBoy has a Z80-like processor. There are lots of Z80
 * emulators but i've done this from zero. All opcodes are coded inside an
 * array of functions. The opcode number is the array index.
 *
 * Yes, it's very awkward, it uses global variables as registers and other
 * ugly programming techniques but it's faster than creating objects or 
 * complex data structures. It also makes the less function calls possible so
 * it's full of repeated code. 
 *
 * Have fun!
 *
 */

var RA=0; // Accumulator
var FZ=0, // bit 7 - Zero
    FN=0, // bit 6 - 
    FH=0, // bit 5 - Half Carry
    FC=0; // bit 4 - Carry
var RB=0; // Register B
var RC=0; // Register C
var RD=0; // Register D
var RE=0; // Register E
var HL=0; // Registers H and L
var SP=0; // Stack Pointer
var PC=0; // Program Counter

var IME=true;   // Interrupt master enable
var DCC=0;      // DIV Cicles Count
var ICC=0;      // Instruction Cicles Count
var TCC=0;      // Timer Cicles Count
var TMC=0;      // Timer Max Cicles (set in MEMW)
var STOP=false; // Used for stopping the CPU (STOP and HALT opcodes)

var OP= new Array; // Opcode Array
var MN= new Array; // Mnemonics

for (var i=0;i<=0xFF;i++) {
  OP[i]=function(){
    pause();
    alert('Unknown opcode: PC='+hex(PC)+' - OP=0x'+hex(MEM[PC]));
  };
}
for (var i=0;i<=0xFF;i++) {
  MN[i]=function(){return 'DB 0x'+hex2(MEM[PC])+'; unknown';};
}

//LD nn,n
OP[0x06]=function(){RB=MEM[++PC];ICC=8;};                     
OP[0x0E]=function(){RC=MEM[++PC];ICC=8;};                     
OP[0x16]=function(){RD=MEM[++PC];ICC=8;};                     
OP[0x1E]=function(){RE=MEM[++PC];ICC=8;};                     
OP[0x26]=function(){HL&=0x00FF;HL|=MEM[++PC]<<8;ICC=8;};
OP[0x2E]=function(){HL&=0xFF00;HL|=MEM[++PC];ICC=8;};
//LD r1,r2
OP[0x7F]=function(){ICC=4;};
OP[0x78]=function(){RA=RB;ICC=4;};                            
OP[0x79]=function(){RA=RC;ICC=4;};                            
OP[0x7A]=function(){RA=RD;ICC=4;};                            
OP[0x7B]=function(){RA=RE;ICC=4;};                            
OP[0x7C]=function(){RA=HL>>8;ICC=4;};                         
OP[0x7D]=function(){RA=HL&0xFF;ICC=4;};                       
OP[0x7E]=function(){RA=MEM[HL];ICC=8;};                       
OP[0x40]=function(){ICC=4;};                                  
OP[0x41]=function(){RB=RC;ICC=4;};                            
OP[0x42]=function(){RB=RD;ICC=4;};                            
OP[0x43]=function(){RB=RE;ICC=4;};                            
OP[0x44]=function(){RB=HL>>8;ICC=4;};                         
OP[0x45]=function(){RB=HL&0xFF;ICC=4;};                       
OP[0x46]=function(){RB=MEM[HL];ICC=8;};                       
OP[0x48]=function(){RC=RB;ICC=4;};                            
OP[0x49]=function(){ICC=4;};                                  
OP[0x4A]=function(){RC=RD;ICC=4;};                            
OP[0x4B]=function(){RC=RE;ICC=4;};                            
OP[0x4C]=function(){RC=HL>>8;ICC=4;};                         
OP[0x4D]=function(){RC=HL&0xFF;ICC=4;};                       
OP[0x4E]=function(){RC=MEM[HL];ICC=8;};                       
OP[0x50]=function(){RD=RB;ICC=4;};                            
OP[0x51]=function(){RD=RC;ICC=4;};                            
OP[0x52]=function(){ICC=4;};                                  
OP[0x53]=function(){RD=RE;ICC=4;};                            
OP[0x54]=function(){RD=HL>>8;ICC=4;};                         
OP[0x55]=function(){RD=HL&0xFF;ICC=4;};                       
OP[0x56]=function(){RD=MEM[HL];ICC=8;};                       
OP[0x58]=function(){RE=RB;ICC=4;};                            
OP[0x59]=function(){RE=RC;ICC=4;};                            
OP[0x5A]=function(){RE=RD;ICC=4;};                            
OP[0x5B]=function(){ICC=4;};                                  
OP[0x5C]=function(){RE=HL>>8;ICC=4;};                            
OP[0x5D]=function(){RE=HL&0xFF;ICC=4;};                            
OP[0x5E]=function(){RE=MEM[HL];ICC=8;};                       
OP[0x60]=function(){HL=(HL&0x00FF)|(RB<<8);ICC=4;};           
OP[0x61]=function(){HL=(HL&0x00FF)|(RC<<8);ICC=4;};           
OP[0x62]=function(){HL=(HL&0x00FF)|(RD<<8);ICC=4;};           
OP[0x63]=function(){HL=(HL&0x00FF)|(RE<<8);ICC=4;};           
OP[0x64]=function(){ICC=4;};                                  
OP[0x65]=function(){HL=(HL&0x00FF)|((HL&0xFF)<<8);ICC=4;};    
OP[0x66]=function(){HL=(HL&0x00FF)|(MEM[HL]<<8);ICC=8;};      
OP[0x68]=function(){HL=(HL&0xFF00)|RB;ICC=4;};                
OP[0x69]=function(){HL=(HL&0xFF00)|RC;ICC=4;};                
OP[0x6A]=function(){HL=(HL&0xFF00)|RD;ICC=4;};                
OP[0x6B]=function(){HL=(HL&0xFF00)|RE;ICC=4;};                
OP[0x6C]=function(){HL=(HL&0xFF00)|(HL>>8);ICC=4;};           
OP[0x6D]=function(){ICC=4;};                                  
OP[0x6E]=function(){HL=(HL&0xFF00)|(MEM[HL]);ICC=8;};         
OP[0x70]=function(){MEMW(HL,RB);ICC=8;};                       
OP[0x71]=function(){MEMW(HL,RC);ICC=8;};                       
OP[0x72]=function(){MEMW(HL,RD);ICC=8;};                       
OP[0x73]=function(){MEMW(HL,RE);ICC=8;};                       
OP[0x74]=function(){MEMW(HL,HL>>8);ICC=8;};                    
OP[0x75]=function(){MEMW(HL,HL&0x00FF);ICC=8;};                
OP[0x36]=function(){MEMW(HL,MEM[++PC]);ICC=12;};                
// LD A,n
OP[0x0A]=function(){RA=MEM[((RB&0x00FF)<<8)|RC];ICC=8;};      
OP[0x1A]=function(){RA=MEM[((RD&0x00FF)<<8)|RE];ICC=8;};      
OP[0x7E]=function(){RA=MEM[HL];ICC=8;};                        
OP[0xFA]=function(){RA=MEM[(MEM[PC+2]<<8)|MEM[PC+1]];PC+=2;ICC=16;}; 
OP[0x3E]=function(){RA=MEM[++PC];ICC=8;};                     
// LD n,A
OP[0x47]=function(){RB=RA;ICC=4;};                            
OP[0x4F]=function(){RC=RA;ICC=4;};                            
OP[0x57]=function(){RD=RA;ICC=4;};                            
OP[0x5F]=function(){RE=RA;ICC=4;};                            
OP[0x67]=function(){HL=(RA<<8)|(HL&0xFF);ICC=4;};                            
OP[0x6F]=function(){HL=RA|(HL&0xFF00);ICC=4;};                            
OP[0x02]=function(){MEMW((RB<<8)|RC,RA);ICC=8;};               
OP[0x12]=function(){MEMW((RD<<8)|RE,RA);ICC=8;};               
OP[0x77]=function(){MEMW(HL,RA);ICC=8;};                       
OP[0xEA]=function(){MEMW((MEM[PC+2]<<8)|MEM[PC+1],RA);PC+=2;ICC=16;}; 
// LD A,(C) = LD A,(0xFF00+C)
OP[0xF2]=function(){RA=MEM[0xFF00+RC];ICC=8;};                
// LD (C),A = LD (0xFF00+C),A
OP[0xE2]=function(){MEMW(0xFF00+RC,RA);ICC=8;};                
// LDD A,(HL) = LD A,(HL-) = LD A,(HLD)
OP[0x3A]=function(){RA=MEM[HL];HL=(HL-1)&0xFFFF;ICC=8;};                     
// LDD (HL),A = LD (HL-),A = LD (HLD),A
OP[0x32]=function(){MEMW(HL,RA);HL=(HL-1)&0xFFFF;ICC=8;};                     
// LDI A,(HL) = LD A,(HL+) = LD A,(HLI)
OP[0x2A]=function(){RA=MEM[HL];HL=(HL+1)&0xFFFF;ICC=8;};                     
// LDI (HL),A = LD (HL+),A = LD (HLI),A
OP[0x22]=function(){MEMW(HL,RA);HL=(HL+1)&0xFFFF;ICC=8;};                     
// LDH (0xFF00+n),A
OP[0xE0]=function(){MEMW(0xFF00+MEM[++PC],RA);ICC=12;};         
// LDH A,(0xFF00+n)
OP[0xF0]=function(){RA=MEM[0xFF00+MEM[++PC]];ICC=12;};         
// LD n,nn
OP[0x01]=function(){RC=MEM[++PC];RB=MEM[++PC];ICC=12;};        
OP[0x11]=function(){RE=MEM[++PC];RD=MEM[++PC];ICC=12;};        
OP[0x21]=function(){HL=(MEM[PC+2]<<8)|MEM[PC+1];PC+=2;ICC=12;};      
OP[0x31]=function(){SP=(MEM[PC+2]<<8)|MEM[PC+1];PC+=2;ICC=12;};      
// LD SP,HL
OP[0xF9]=function(){SP=HL;ICC=8;};                            
// LDHL SP,n = LD HL,SP+n TODO: flags are incorrect?
OP[0xF8]=function(){var n=MEM[++PC];HL=SP+sb(n);FZ=0;RN=0;
  FH=(((SP&0x0F)+(n&0x0F))>0x0F);
  FC=(((SP&0xFF)+(n&0xFF))>0xFF);
  ICC=12;};                 
// LD (nn),SP
OP[0x08]=function(){MEMW((MEM[++PC]<<8)+MEM[++PC],SP);ICC=20;}; 
// PUSH nn
OP[0xF5]=function(){MEMW(--SP,RA);MEMW(--SP,(FZ<<7)|(FN<<6)|(FH<<5)|(FC<<4));ICC=16;};          
OP[0xC5]=function(){MEMW(--SP,RB);MEMW(--SP,RC);ICC=16;};
OP[0xD5]=function(){MEMW(--SP,RD);MEMW(--SP,RE);ICC=16;};
OP[0xE5]=function(){MEMW(--SP,HL>>8);MEMW(--SP,HL&0xFF);ICC=16;};
// POP nn
OP[0xF1]=function(){var FR=MEM[SP++];RA=MEM[SP++];
  FZ=(FR>>7)&1;FN=(FR>>6)&1;
  FH=(FR>>5)&1;FC=(FR>>4)&1;
  ICC=12;};                      
MN[0xF1]=function(){return 'POP AF';};

OP[0xC1]=function(){RC=MEM[SP++];RB=MEM[SP++];ICC=12;};
MN[0xC1]=function(){return 'POP BC';};

OP[0xD1]=function(){RE=MEM[SP++];RD=MEM[SP++];ICC=12;}; 
MN[0xD1]=function(){return 'POP DE';};

OP[0xE1]=function(){var L=MEM[SP++]; HL=(MEM[SP++]<<8)|L;ICC=12;};      
MN[0xE1]=function(){return 'POP HL';};

// ADD A,n
OP[0x87]=function(){
  FH=((RA&0x0F)*2)>0x0F;
  FC=((RA&0xFF)*2)>0xFF;
  RA=(RA*2)&0xFF;FZ=RA==0;FN=0;ICC=4;};                 
MN[0x87]=function(){return 'ADD A,A';};

OP[0x80]=function(){
  FH=((RA&0x0F)+(RB&0x0F))>0x0F;
  FC=((RA&0xFF)+(RB&0xFF))>0xFF;
  RA=(RA+RB)&0xFF;FZ=RA==0;FN=0;ICC=4;};                
MN[0x80]=function(){return 'ADD A,B';};

OP[0x81]=function(){
  FH=((RA&0x0F)+(RC&0x0F))>0x0F;
  FC=((RA&0xFF)+(RC&0xFF))>0xFF;
  RA=(RA+RC)&0xFF;FZ=RA==0;FN=0;ICC=4;};                
MN[0x81]=function(){return 'ADD A,C';};

OP[0x82]=function(){
  FH=((RA&0x0F)+(RD&0x0F))>0x0F;
  FC=((RA&0xFF)+(RD&0xFF))>0xFF;
  RA=(RA+RD)&0xFF;FZ=RA==0;FN=0;ICC=4;};                
MN[0x82]=function(){return 'ADD A,D';};

OP[0x83]=function(){
  FH=((RA&0x0F)+(RE&0x0F))>0x0F;
  FC=((RA&0xFF)+(RE&0xFF))>0xFF;
  RA=(RA+RE)&0xFF;FZ=RA==0;FN=0;};                
MN[0x83]=function(){return 'ADD A,E';};

OP[0x84]=function(){ var H=HL>>8;
  FH=((RA&0x0F)+(H&0x0F))>0x0F;
  FC=((RA&0xFF)+(H&0xFF))>0xFF;
  RA=(RA+H)&0xFF;FZ=RA==0;FN=0;ICC=4;};                 
MN[0x84]=function(){return 'ADD A,H';};

OP[0x85]=function(){ var L=HL&0xFF;
  FH=((RA&0x0F)+(L&0x0F))>0x0F;
  FC=((RA&0xFF)+(L&0xFF))>0xFF;
  RA=(RA+L)&0xFF;FZ=RA==0;FN=0;ICC=4;};                 
MN[0x85]=function(){return 'ADD A,L';};

OP[0x86]=function(){ var M=MEM[HL];
  FH=((RA&0x0F)+(M&0x0F))>0x0F;
  FC=((RA&0xFF)+(M&0xFF))>0xFF;
  RA=(RA+M)&0xFF;FZ=RA==0;FN=0;ICC=8;};                 
MN[0x86]=function(){return 'ADD A,(HL)';};

OP[0xC6]=function(){ var M=MEM[++PC];
  FH=((RA&0x0F)+(M&0x0F))>0x0F;
  FC=((RA&0xFF)+(M&0xFF))>0xFF;
  RA=(RA+M)&0xFF;FZ=RA==0;FN=0;ICC=8;};                 
MN[0xC6]=function(){return 'ADD A,0x'+hex2(MEM[PC+1]);};

// ADC A,n
OP[0x8F]=function(){var C=FC;
  FH=((RA&0x0F)+(RA&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(RA&0xFF)+C)>0xFF;
  RA=(RA+RA+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};              
MN[0x8F]=function(){return 'ADC A,A';};

OP[0x88]=function(){var C=FC;
  FH=((RA&0x0F)+(RB&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(RB&0xFF)+C)>0xFF;
  RA=(RA+RB+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};              
MN[0x88]=function(){return 'ADC A,B';};

OP[0x89]=function(){var C=FC;
  FH=((RA&0x0F)+(RC&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(RC&0xFF)+C)>0xFF;
  RA=(RA+RC+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};              
MN[0x89]=function(){return 'ADC A,C';};

OP[0x8A]=function(){var C=FC;
  FH=((RA&0x0F)+(RD&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(RD&0xFF)+C)>0xFF;
  RA=(RA+RD+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};              
MN[0x8A]=function(){return 'ADC A,D';};

OP[0x8B]=function(){var C=FC;
  FH=((RA&0x0F)+(RE&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(RE&0xFF)+C)>0xFF;
  RA=(RA+RE+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};              
MN[0x8B]=function(){return 'ADC A,E';};

OP[0x8C]=function(){var H=HL>>8;var C=FC;
  FH=((RA&0x0F)+(H&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(H&0xFF)+C)>0xFF;
  RA=(RA+H+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};               
MN[0x8C]=function(){return 'ADC A,H';};

OP[0x8D]=function(){var L=HL&0xFF;var C=FC;
  FH=((RA&0x0F)+(L&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(L&0xFF)+C)>0xFF;
  RA=(RA+L+C)&0xFF;FZ=RA==0;FN=0;ICC=4;};               
MN[0x8E]=function(){return 'ADC A,L';};

OP[0x8E]=function(){var M=MEM[HL];var C=FC;
  FH=((RA&0x0F)+(M&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(M&0xFF)+C)>0xFF;
  RA=(RA+M+C)&0xFF;FZ=RA==0;FN=0;ICC=8;};               
MN[0x8E]=function(){return 'ADC A,(HL)';};

OP[0xCE]=function(){var M=MEM[++PC];var C=FC;
  FH=((RA&0x0F)+(M&0x0F)+C)>0x0F;
  FC=((RA&0xFF)+(M&0xFF)+C)>0xFF;
  RA=(RA+M+C)&0xFF;FZ=RA==0;FN=0;ICC=8;};               
MN[0xCE]=function(){return 'ADC A,0x'+hex2(MEM[PC+1]);};

// SUB n
OP[0x97]=function(){FH=0;FC=0;RA=0;FZ=1;FN=1;ICC=4;};          
MN[0x97]=function(){return 'SUB A';};

OP[0x90]=function(){
  FH=((RA&0x0F)-(RB&0x0F))<0;
  FC=((RA&0xFF)-(RB&0xFF))<0;
  RA=(RA-RB)&0xFF;FZ=RA==0;FN=1;ICC=4;};                
MN[0x90]=function(){return 'SUB B';};

OP[0x91]=function(){
  FH=((RA&0x0F)-(RC&0x0F))<0;
  FC=((RA&0xFF)-(RC&0xFF))<0;
  RA=(RA-RC)&0xFF;FZ=RA==0;FN=1;ICC=4;};                
MN[0x91]=function(){return 'SUB C';};

OP[0x92]=function(){
  FH=((RA&0x0F)-(RD&0x0F))<0;
  FC=((RA&0xFF)-(RD&0xFF))<0;
  RA=(RA-RD)&0xFF;FZ=RA==0;FN=1;ICC=4;};                
MN[0x92]=function(){return 'SUB D';};

OP[0x93]=function(){
  FH=((RA&0x0F)-(RE&0x0F))<0;
  FC=((RA&0xFF)-(RE&0xFF))<0;
  RA=(RA-RE)&0xFF;FZ=RA==0;FN=1;ICC=4;};                
MN[0x93]=function(){return 'SUB E';};

OP[0x94]=function(){var H=HL>>8;
  FH=((RA&0x0F)-(H&0x0F))<0;
  FC=((RA&0xFF)-(H&0xFF))<0;
  RA=(RA-H)&0xFF;FZ=RA==0;FN=1;ICC=4;};                 
MN[0x94]=function(){return 'SUB H';};

OP[0x95]=function(){var L=HL&0xFF;
  FH=((RA&0x0F)-(L&0x0F))<0;
  FC=((RA&0xFF)-(L&0xFF))<0;
  RA=(RA-L)&0xFF;FZ=RA==0;FN=1;ICC=4;};                 
MN[0x95]=function(){return 'SUB L';};

OP[0x96]=function(){var M=MEM[HL];
  FH=((RA&0x0F)-(M&0x0F))<0;
  FC=((RA&0xFF)-(M&0xFF))<0;
  RA=(RA-M)&0xFF;FZ=RA==0;FN=1;ICC=8;};                 
MN[0x96]=function(){return 'SUB C';};

OP[0xD6]=function(){var M=MEM[++PC];
  FH=((RA&0x0F)-(M&0x0F))<0;
  FC=((RA&0xFF)-(M&0xFF))<0;
  RA=(RA-M)&0xFF;FZ=RA==0;FN=1;ICC=8;};                 
MN[0xD6]=function(){return 'SUB 0x'+hex2(MEM[PC+1]);};

//SBC A,n
// TODO: SBC A,# -> unknown opcode
// TODO: check flags
OP[0x9F]=function(){var C=FC;
  FH=C>0;FC=C>0;
  RA=(-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};                   
MN[0x9F]=function(){return 'SBC A,A';};

OP[0x98]=function(){var C=FC;
  FH=((RA&0x0F)-(RB&0x0F)-C)<0;
  FC=((RA&0xFF)-(RB&0xFF)-C)<0;
  RA=(RA-RB-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};              
MN[0x98]=function(){return 'SBC A,B';};

OP[0x99]=function(){var C=FC;
  FH=((RA&0x0F)-(RC&0x0F)-C)<0;
  FC=((RA&0xFF)-(RC&0xFF)-C)<0;
  RA=(RA-RC-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};              
MN[0x99]=function(){return 'SBC A,C';};

OP[0x9A]=function(){var C=FC;
  FH=((RA&0x0F)-(RD&0x0F)-C)<0;
  FC=((RA&0xFF)-(RD&0xFF)-C)<0;
  RA=(RA-RD-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};              
MN[0x9A]=function(){return 'SBC A,D';};

OP[0x9B]=function(){var C=FC;
  FH=((RA&0x0F)-(RE&0x0F)-C)<0;
  FC=((RA&0xFF)-(RE&0xFF)-C)<0;
  RA=(RA-RE-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};              
MN[0x9B]=function(){return 'SBC A,E';};

OP[0x9C]=function(){var C=FC;var H=HL>>8;
  FH=((RA&0x0F)-(H&0x0F)-C)<0;
  FC=((RA&0xFF)-(H&0xFF)-C)<0;
  RA=(RA-H-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};               
MN[0x9C]=function(){return 'SBC A,H';};

OP[0x9D]=function(){var C=FC;var L=HL&0xFF;
  FH=((RA&0x0F)-(L&0x0F)-C)<0;
  FC=((RA&0xFF)-(L&0xFF)-C)<0;
  RA=(RA-L-C)&0xFF;FZ=RA==0;FN=1;ICC=4;};               
MN[0x9D]=function(){return 'SBC A,L';};

OP[0x9E]=function(){var C=FC;var M=MEM[HL];
  FH=((RA&0x0F)-(M&0x0F)-C)<0;
  FC=((RA&0xFF)-(M&0xFF)-C)<0;
  RA=(RA-M-C)&0xFF;FZ=RA==0;FN=1;ICC=8;};               
MN[0x9E]=function(){return 'SBC A,(HL)';};

OP[0xDE]=function(){var C=FC;var V=MEM[++PC];
  FH=((RA&0x0F)-(V&0x0F)-C)<0;
  FC=((RA&0xFF)-(V&0xFF)-C)<0;
  RA=(RA-V-C)&0xFF;FZ=RA==0;FN=1;ICC=8;};               
MN[0xDE]=function(){return 'SBC A,0x'+hex2(MEM[PC+1]);};

// AND n
OP[0xA7]=function(){FZ=RA==0;
  FN=0;FH=1;FC=0;ICC=4;};    
MN[0xA7]=function(){return 'AND A';};

OP[0xA0]=function(){RA&=RB;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA0]=function(){return 'AND B';};

OP[0xA1]=function(){RA&=RC;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA1]=function(){return 'AND C';};

OP[0xA2]=function(){RA&=RD;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA2]=function(){return 'AND D';};

OP[0xA3]=function(){RA&=RE;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA3]=function(){return 'AND E';};

OP[0xA4]=function(){RA&=HL>>8;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA4]=function(){return 'AND H';};

OP[0xA5]=function(){RA&=HL&0xFF;
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=4;};                      
MN[0xA5]=function(){return 'AND L';};

OP[0xA6]=function(){RA&=MEM[HL];
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=8;};                      
MN[0xA6]=function(){return 'AND (HL)';};

OP[0xE6]=function(){RA&=MEM[++PC];
  FZ=RA==0;FN=0;FH=1;FC=0;ICC=8;};                      
MN[0xE6]=function(){return 'AND 0x'+hex2(MEM[PC+1]);};

// OR n
OP[0xB7]=function(){FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};    
MN[0xB7]=function(){return 'OR A';};

OP[0xB0]=function(){RA|=RB;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB0]=function(){return 'OR B';};

OP[0xB1]=function(){RA|=RC;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB1]=function(){return 'OR C';};

OP[0xB2]=function(){RA|=RD;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB2]=function(){return 'OR D';};

OP[0xB3]=function(){RA|=RE;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB3]=function(){return 'OR E';};

OP[0xB4]=function(){RA|=HL>>8;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB4]=function(){return 'OR H';};

OP[0xB5]=function(){RA|=HL&0xFF;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xB5]=function(){return 'OR L';};

OP[0xB6]=function(){RA|=MEM[HL];
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=8;};                      
MN[0xB6]=function(){return 'OR (HL)';};

OP[0xF6]=function(){RA|=MEM[++PC];
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=8;};                      
MN[0xF6]=function(){return 'OR 0x'+hex2(MEM[PC+1]);};

// XOR n
OP[0xAF]=function(){RA=0;FZ=1;FN=0;FH=0;FC=0;ICC=4;};         
MN[0xAF]=function(){return 'XOR A';};

OP[0xA8]=function(){RA^=RB;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xA8]=function(){return 'XOR B';};

OP[0xA9]=function(){RA^=RC;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xA9]=function(){return 'XOR C';};

OP[0xAA]=function(){RA^=RD;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xAA]=function(){return 'XOR D';};

OP[0xAB]=function(){RA^=RE;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xAB]=function(){return 'XOR E';};

OP[0xAC]=function(){RA^=HL>>8;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xAC]=function(){return 'XOR H';};

OP[0xAD]=function(){RA^=HL&0xFF;
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=4;};                      
MN[0xAD]=function(){return 'XOR L';};

OP[0xAE]=function(){RA^=MEM[HL];
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=8;};                      
MN[0xAE]=function(){return 'XOR (HL)';};

OP[0xEE]=function(){RA^=MEM[++PC];
  FZ=RA==0;FN=0;FH=0;FC=0;ICC=8;};                      
MN[0xEE]=function(){return 'XOR 0x'+hex2(MEM[PC+1]);};

// CP n
// TODO: test flags
OP[0xBF]=function(){FZ=1;FN=1;FC=0;FH=0;ICC=4;};              
MN[0xBF]=function(){return 'CP A';};

OP[0xB8]=function(){FZ=RA==RB;FN=1;
  FC=RA<RB;
  FH=(RA&0x0F)<(RB&0x0F);ICC=4;};          
MN[0xB8]=function(){return 'CP B';};

OP[0xB9]=function(){FZ=RA==RC;FN=1;
  FC=RA<RC;
  FH=(RA&0x0F)<(RC&0x0F);ICC=4;};          
MN[0xB9]=function(){return 'CP C';};

OP[0xBA]=function(){FZ=RA==RD;FN=1;
  FC=RA<RD;
  FH=(RA&0x0F)<(RD&0x0F);ICC=4;};          
MN[0xBA]=function(){return 'CP D';};

OP[0xBB]=function(){FZ=RA==RE;FN=1;
  FC=RA<RE;
  FH=(RA&0x0F)<(RE&0x0F);ICC=4;};         
MN[0xBB]=function(){return 'CP E';};

OP[0xBC]=function(){var H=HL>>8;FZ=RA==H;FN=1;
  FC=RA<H;
  FH=(RA&0x0F)<(H&0x0F);ICC=4;};            
MN[0xBC]=function(){return 'CP H';};

OP[0xBD]=function(){var L=HL&0xFF;FZ=RA==L;FN=1;
  FC=RA<L;
  FH=(RA&0x0F)<(L&0x0F);ICC=4;};            
MN[0xBD]=function(){return 'CP L';};

OP[0xBE]=function(){var M=MEM[HL];FZ=RA==M;FN=1;
  FC=RA<M;
  FH=(RA&0x0F)<(M&0x0F);ICC=8;};            
MN[0xBE]=function(){return 'CP (HL)';};

OP[0xFE]=function(){var M=MEM[++PC];FZ=RA==M;FN=1;
  FC=RA<M;
  FH=(RA&0x0F)<(M&0x0F);ICC=8;};            
MN[0xFE]=function(){return 'CP '+MEM[PC+1]+'; 0x'+hex2(MEM[PC+1]);};

// INC n
OP[0x3C]=function(){RA=(++RA)&0xFF;
  FZ=RA==0; FN=0; FH=(RA&0xF)==0;ICC=4;};         
MN[0x3C]=function(){return 'INC A';};

OP[0x04]=function(){RB=(++RB)&0xFF;
  FZ=RB==0; FN=0; FH=(RB&0xF)==0;ICC=4;};         
MN[0x04]=function(){return 'INC B';};

OP[0x0C]=function(){RC=(++RC)&0xFF;
  FZ=RC==0; FN=0; FH=(RC&0xF)==0;ICC=4;};         
MN[0x0C]=function(){return 'INC C';};

OP[0x14]=function(){RD=(++RD)&0xFF;
  FZ=RD==0; FN=0; FH=(RD&0xF)==0;ICC=4;};         
MN[0x14]=function(){return 'INC D';};

OP[0x1C]=function(){RE=(++RE)&0xFF;
  FZ=RE==0; FN=0; FH=(RE&0xF)==0;ICC=4;};         
MN[0x1C]=function(){return 'INC E';};

OP[0x24]=function(){var H=((HL>>8)+1)&0xFF;
  HL=(HL&0x00FF)|(H<<8);
  FZ=H==0; FN=0; FH=(H&0xF)==0;ICC=4;};           
MN[0x24]=function(){return 'INC H';};

OP[0x2C]=function(){var L=(HL+1)&0xFF;
  HL=(HL&0xFF00)|L;
  FZ=L==0; FN=0; FH=(L&0xF)==0;ICC=4;};           
MN[0x2C]=function(){return 'INC L';};

OP[0x34]=function(){var M=(MEM[HL]+1)&0xFF;
  MEM[HL]=M;FZ=M==0;
  FN=0; FH=(M&0xF)==0; MEM[HL]=M;ICC=12;};               
MN[0x34]=function(){return 'INC (HL)';};

// DEC n
// TODO test flags
OP[0x3D]=function(){
  RA=(RA-1)&0xFF;
  FZ=RA==0;
  FN=1;
  FH=(RA&0xF)==0xF;ICC=4;};         
MN[0x3D]=function(){return 'DEC A';};

OP[0x05]=function(){
  RB=(RB-1)&0xFF;
  FZ=RB==0;
  FN=1;
  FH=(RB&0xF)==0xF;ICC=4;};         
MN[0x05]=function(){return 'DEC B';};

OP[0x0D]=function(){
  RC=(RC-1)&0xFF;
  FZ=RC==0;
  FN=1;
  FH=(RC&0xF)==0xF;ICC=4;};         
MN[0x0D]=function(){return 'DEC C';};

OP[0x15]=function(){
  RD=(RD-1)&0xFF;
  FZ=RD==0;
  FN=1;
  FH=(RD&0xF)==0xF;ICC=4;};         
MN[0x15]=function(){return 'DEC D';};

OP[0x1D]=function(){
  RE=(RE-1)&0xFF;
  FZ=RE==0;
  FN=1;
  FH=(RE&0xF)==0xF;ICC=4;};         
MN[0x1D]=function(){return 'DEC E';};

OP[0x25]=function(){var H=((HL>>8)-1)&0xFF;
  HL=(HL&0x00FF)|(H<<8);
  FZ=H==0;
  FN=1;
  FH=(H&0xF)==0xF;ICC=4;};                         
MN[0x25]=function(){return 'DEC H';};

OP[0x2D]=function(){var L=(HL+1)&0xFF;
  HL=(HL&0xFF00)|L;
  FZ=L==0;
  FN=1;
  FH=(L&0xF)==0xF;ICC=4;};                         
MN[0x2D]=function(){return 'DEC L';};

OP[0x35]=function(){var M=(MEM[HL]-1)&0xFF;
  MEM[HL]=M;
  FZ=M==0;
  FN=1;
  FH=(M&0xF)==0xF;ICC=12;};                         
MN[0x35]=function(){return 'DEC (HL)';};

// ADD HL,n
OP[0x09]=function(){var BC=(RB<<8)|RC;
  FN=0;
  FH=((HL&0xFFF)+(BC&0xFFF))>0xFFF;
  HL+=BC;
  FC=HL>0xFFFF;
  HL&=0xFFFF;ICC=8;};
MN[0x09]=function(){return 'ADD HL,BC';};

OP[0x19]=function(){var DE=(RD<<8)|RE;
  FN=0;
  FH=((HL&0xFFF)+(DE&0xFFF))>0xFFF;
  HL+=DE;
  FC=HL>0xFFFF;
  HL&=0xFFFF;ICC=8;};
MN[0x19]=function(){return 'ADD HL,DE';};

OP[0x29]=function(){
  FN=0;
  FH=((HL&0xFFF)*2)>0xFFF;
  HL+=HL;
  FC=HL>0xFFFF;
  HL&=0xFFFF;ICC=8;};
MN[0x29]=function(){return 'ADD HL,HL';};

OP[0x39]=function(){
  FN=0;
  FH=((HL&0xFFF)+(SP&0xFFF))>0xFFF;
  HL+=SP;
  FC=HL>0xFFFF;
  HL&=0xFFFF;ICC=8;};
MN[0x39]=function(){return 'ADD HL,SP';};

//ADD SP,n
OP[0xE8]=function(){var n=MEM[++PC];
  FZ=0;
  FN=0;
  FH=((SP&0xFFF)+n)>0xFFF; // test bit 11. I'm unsure on this.
  SP+=sb(n);
  FC=SP>0xFFFF;
  SP&=0xFFFF;ICC=16;};
MN[0xE8]=function(){return 'ADD SP,0x'+hex2(MEM[PC+1]);};

//INC nn
OP[0x03]=function(){var BC=((RB<<8)+RC+1)&0xFFFF;
  RB=BC>>8;
  RC=BC&0xFF;ICC=8;};
MN[0x03]=function(){return 'INC BC';};

OP[0x13]=function(){var DE=((RD<<8)+RE+1)&0xFFFF;
  RD=DE>>8;
  RE=DE&0xFF;ICC=8;};
MN[0x13]=function(){return 'INC DE';};

OP[0x23]=function(){HL=(HL+1)&0xFFFF;ICC=8;};
MN[0x23]=function(){return 'INC HL';};

OP[0x33]=function(){SP=(SP+1)&0xFFFF;ICC=8;};
MN[0x33]=function(){return 'INC SP';};

//DEC nn
OP[0x0B]=function(){var BC=((RB<<8)+RC-1)&0xFFFF;
  RB=BC>>8;
  RC=BC&0xFF;ICC=8;};
MN[0x0B]=function(){return 'DEC BC';};

OP[0x1B]=function(){var DE=((RD<<8)+RE-1)&0xFFFF;
  RD=DE>>8;
  RE=DE&0xFF;ICC=8;};
MN[0x1B]=function(){return 'DEC DE';};

OP[0x2B]=function(){HL=(HL-1)&0xFFFF;ICC=8;};
MN[0x2B]=function(){return 'DEC HL';};

OP[0x3B]=function(){SP=(SP-1)&0xFFFF;ICC=8;};
MN[0x3B]=function(){return 'DEC SP';};

// DDA
OP[0x27]=function(){ICC=4; /* TODO implement */ };
MN[0x27]=function(){return 'DDA';};

//CPL
OP[0x2F]=function(){RA=(~RA)&0xFF;FN=1;FH=1;ICC=4;};
MN[0x2F]=function(){return 'CPL';};

//CCF
OP[0x3F]=function(){FC=(~FC)&1;FN=0;FH=0;ICC=4;};
MN[0x3F]=function(){return 'CCF';};

//SCF
OP[0x37]=function(){FC=1;FN=0;FH=0;ICC=4;};
MN[0x37]=function(){return 'SCF';};

//NOP
OP[0x00]=function(){ICC=4;};
MN[0x00]=function(){return 'NOP';};

//HALT
OP[0x76]=function(){ICC=4;}; // TODO 
MN[0x76]=function(){return 'HALT';};

//STOP
OP[0x10]=function(){ICC=4;PC++;}; // TODO
MN[0x10]=function(){return 'STOP'};

//DI
OP[0xF3]=function(){IME=false;ICC=4;}; // TODO 
MN[0xF3]=function(){return 'DI';};

//EI
OP[0xFB]=function(){IME=true;ICC=4;}; // TODO 
MN[0xFB]=function(){return 'EI';};

//RLCA
OP[0x07]=function(){
  FC=(RA>>7)&1;
  RA=((RA<<1)&0xFF)|FC;
  FN=0;
  FH=0;
  FZ=RA==0;ICC=4;}; 
MN[0x07]=function(){return 'RLCA';};

//RLA
OP[0x17]=function(){var C=FC;
  FC=(RA>>7)&1;
  RA=((RA<<1)&0xFF)|C;
  FN=0;
  FH=0;
  FZ=RA==0;ICC=4;}; 
MN[0x17]=function(){return 'RLA';};

//RRCA
OP[0x0F]=function(){
  FC=RA&1;
  RA=(RA>>1)|(FC<<7);
  FN=0;
  FH=0;
  FZ=RA==0;ICC=4;}; 
MN[0x0F]=function(){return 'RRCA';};

//RRA
OP[0x1F]=function(){var C=FC;
  FC=RA&1;
  RA=(RA>>1)|(C<<7);
  FN=0;
  FH=0;
  FZ=RA==0;ICC=4;}; 
MN[0x1F]=function(){return 'RRA';};

OPCB=new Array(0x100);
MNCB=new Array(0x100);
for (var i=0;i<=0xFF;i++){
  OPCB[i]=function(){};
  MNCB[i]=function(){return 'DW 0xCB'+hex2(MEM[PC+1]);};
}
OP[0xCB]=function(){OPCB[MEM[++PC]]();};
MN[0xCB]=function(){return MNCB[MEM[PC+1]]();};

//SWAP n
OPCB[0x37]=function(){RA=((RA<<4)|(RA>>4))&0xFF;ICC=8;};
MNCB[0x37]=function(){return 'SWAP A';};

OPCB[0x30]=function(){RB=((RB<<4)|(RB>>4))&0xFF;ICC=8;};
MNCB[0x30]=function(){return 'SWAP B';};

OPCB[0x31]=function(){RC=((RC<<4)|(RC>>4))&0xFF;ICC=8;};
MNCB[0x31]=function(){return 'SWAP C';};

OPCB[0x32]=function(){RD=((RD<<4)|(RD>>4))&0xFF;ICC=8;};
MNCB[0x32]=function(){return 'SWAP D';};

OPCB[0x33]=function(){RE=((RE<<4)|(RE>>4))&0xFF;ICC=8;};
MNCB[0x33]=function(){return 'SWAP E';};

OPCB[0x34]=function(){var H=HL>>8;
  H=((H<<4)|(H>>4))&0xFF;
  HL=(HL&0x00FF)|H;ICC=8;};
MNCB[0x34]=function(){return 'SWAP H';};

OPCB[0x35]=function(){var L=HL&0xFF;
  L=((L<<4)|(L>>4))&0xFF;
  HL=(HL&0xFF00)|L;ICC=8;};
MNCB[0x35]=function(){return 'SWAP L';};

OPCB[0x36]=function(){var M=MEM[HL];
  M=((M<<4)|(M>>4))&0xFF;
  MEM[HL]=M;ICC=16;};
MNCB[0x36]=function(){return 'SWAP (HL)';};

//RLC
OPCB[0x07]=function(){
  FC=(RA>>7)&1;
  RA=((RA<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RA==0;ICC=8;};
MNCB[0x07]=function(){return 'RLC A';};

OPCB[0x00]=function(){
  FC=(RB>>7)&1;
  RB=((RB<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RB==0;ICC=8;};
MNCB[0x00]=function(){return 'RLC B';};

OPCB[0x01]=function(){
  FC=(RC>>7)&1;
  RC=((RC<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RC==0;ICC=8;};
MNCB[0x01]=function(){return 'RLC C';};

OPCB[0x02]=function(){
  FC=(RD>>7)&1;
  RD=((RD<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RD==0;ICC=8;};
MNCB[0x02]=function(){return 'RLC D';};

OPCB[0x03]=function(){
  FC=(RE>>7)&1;
  RE=((RE<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RE==0;ICC=8;};
MNCB[0x03]=function(){return 'RLC E';};

OPCB[0x04]=function(){var H=HL>>8;
  FC=(H>>7)&1;
  H=((H<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0xFF);ICC=8;};
MNCB[0x04]=function(){return 'RLC H';};

OPCB[0x05]=function(){var L=HL&0xFF;
  FC=(RL>>7)&1;
  L=((RL<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=RL==0;
  HL=(HL&0xFF00)|L;ICC=8;};
MNCB[0x05]=function(){return 'RLC L';};

OPCB[0x06]=function(){var M=MEM[HL];
  FC=(M>>7)&1;
  M=((M<<1)&0xFF)|FC;
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;};
MNCB[0x06]=function(){return 'RLC (HL)';};

//RL n
OPCB[0x17]=function(){var C=FC;
  FC=(RA>>7)&1;
  RA=((RA<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RA==0;ICC=8;};
MNCB[0x17]=function(){return 'RL A';};

OPCB[0x10]=function(){var C=FC;
  FC=(RB>>7)&1;
  RB=((RB<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RB==0;ICC=8;};
MNCB[0x10]=function(){return 'RL B';};

OPCB[0x11]=function(){var C=FC;
  FC=(RC>>7)&1;
  RC=((RC<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RC==0;ICC=8;};
MNCB[0x11]=function(){return 'RL C';};

OPCB[0x12]=function(){var C=FC;
  FC=(RD>>7)&1;
  RD=((RD<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RD==0;ICC=8;};
MNCB[0x12]=function(){return 'RL D';};

OPCB[0x13]=function(){var C=FC;
  FC=(RE>>7)&1;
  RE=((RE<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RE==0;ICC=8;};
MNCB[0x13]=function(){return 'RL E';};

OPCB[0x14]=function(){var C=FC;var H=HL<<8;
  FC=(RH>>7)&1;
  H=((RH<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RH==0;
  HL=(H<<8)|(HL&0xFF);ICC=8;};
MNCB[0x14]=function(){return 'RL H';};

OPCB[0x15]=function(){var C=FC;var L=HL&0xFF;
  FC=(RL>>7)&1;
  L=((RL<<1)&0xFF)|C;
  FN=0;FH=0;FZ=RL==0;
  HL=(HL&0xFF00)|L;ICC=8;};
MNCB[0x15]=function(){return 'RL L';};

OPCB[0x16]=function(){var C=FC;var M=MEM[HL];
  FC=(M>>7)&1;
  M=((M<<1)&0xFF)|C;
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;};
MNCB[0x16]=function(){return 'RL (HL)';};

//RRC n
OPCB[0x0F]=function(){
  FC=RA&1;
  RA=(RA>>1)|(FC<<7);
  FN=0;FH=0;FZ=RA==0;ICC=8;}; 
MNCB[0x0F]=function(){return 'RRC A';};

OPCB[0x08]=function(){
  FC=RB&1;
  RB=(RB>>1)|(FC<<7);
  FN=0;FH=0;FZ=RB==0;ICC=8;}; 
MNCB[0x08]=function(){return 'RRC B';};

OPCB[0x09]=function(){
  FC=RC&1;
  RC=(RC>>1)|(FC<<7);
  FN=0;FH=0;FZ=RC==0;ICC=8;}; 
MNCB[0x09]=function(){return 'RRC C';};

OPCB[0x0A]=function(){
  FC=RD&1;
  RD=(RD>>1)|(FC<<7);
  FN=0;FH=0;FZ=RD==0;ICC=8;}; 
MNCB[0x0A]=function(){return 'RRC D';};

OPCB[0x0B]=function(){
  FC=RE&1;
  RE=(RE>>1)|(FC<<7);
  FN=0;FH=0;FZ=RE==0;ICC=8;}; 
MNCB[0x0B]=function(){return 'RRC E';};

OPCB[0x0C]=function(){var H=HL>>8;
  FC=H&1;
  H=(H>>1)|(FC<<7);
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0x00FF);ICC=8;}; 
MNCB[0x0C]=function(){return 'RRC H';};

OPCB[0x0D]=function(){var L=HL&0xFF;
  FC=L&1;
  L=(L>>1)|(FC<<7);
  FN=0;FH=0;FZ=L==0;
  HL=(HL&0xFF00)|L;ICC=8;}; 
MNCB[0x0D]=function(){return 'RRC L';};

OPCB[0x0E]=function(){var M=MEM[HL];
  FC=M&1;
  M=(M>>1)|(FC<<7);
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;}; 
MNCB[0x0E]=function(){return 'RRC (HL)';};

//RR n
OPCB[0x1F]=function(){var C=FC;
  FC=RA&1;
  RA=(RA>>1)|(C<<7);
  FN=0;FH=0;FZ=RA==0;ICC=8;}; 
MNCB[0x1F]=function(){return 'RR A';};

OPCB[0x18]=function(){var C=FC;
  FC=RB&1;
  RB=(RB>>1)|(C<<7);
  FN=0;FH=0;FZ=RB==0;ICC=8;}; 
MNCB[0x18]=function(){return 'RR B';};

OPCB[0x19]=function(){var C=FC;
  FC=RC&1;
  RC=(RC>>1)|(C<<7);
  FN=0;FH=0;FZ=RC==0;ICC=8;}; 
MNCB[0x19]=function(){return 'RR C';};

OPCB[0x1A]=function(){var C=FC;
  FC=RD&1;
  RD=(RD>>1)|(C<<7);
  FN=0;FH=0;FZ=RD==0;ICC=8;}; 
MNCB[0x1A]=function(){return 'RR D';};

OPCB[0x1B]=function(){var C=FC;
  FC=RE&1;
  RE=(RE>>1)|(C<<7);
  FN=0;FH=0;FZ=RE==0;ICC=8;}; 
MNCB[0x1B]=function(){return 'RR E';};

OPCB[0x1C]=function(){var C=FC;var H=HL>>8;
  FC=H&1;
  H=(H>>1)|(C<<7);
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0x00FF);ICC=8;}; 
MNCB[0x1C]=function(){return 'RR H';};

OPCB[0x1D]=function(){var C=FC;var L=HL&0xFF;
  FC=L&1;
  L=(L>>1)|(C<<7);
  FN=0;FH=0;FZ=L==0;
  HL=(HL&0xFF00)|L;ICC=8;}; 
MNCB[0x1D]=function(){return 'RR L';};

OPCB[0x1E]=function(){var C=FC;var M=MEM[HL];
  FC=M&1;
  M=(M>>1)|(C<<7);
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;}; 
MNCB[0x1E]=function(){return 'RR (HL)';};

//SLA n
OPCB[0x27]=function(){
  FC=(RA>>7)&1;
  RA=(RA<<1)&0xFF;
  FN=0;FH=0;FZ=RA==0;ICC=8;}; 
MNCB[0x27]=function(){return 'SLA A';};

OPCB[0x20]=function(){
  FC=(RB>>7)&1;
  RB=(RB<<1)&0xFF;
  FN=0;FH=0;FZ=RB==0;ICC=8;}; 
MNCB[0x20]=function(){return 'SLA B';};

OPCB[0x21]=function(){
  FC=(RC>>7)&1;
  RC=(RC<<1)&0xFF;
  FN=0;FH=0;FZ=RC==0;ICC=8;}; 
MNCB[0x21]=function(){return 'SLA C';};

OPCB[0x22]=function(){
  FC=(RD>>7)&1;
  RD=(RD<<1)&0xFF;
  FN=0;FH=0;FZ=RD==0;ICC=8;}; 
MNCB[0x22]=function(){return 'SLA D';};

OPCB[0x23]=function(){
  FC=(RE>>7)&1;
  RE=(RE<<1)&0xFF;
  FN=0;FH=0;FZ=RE==0;ICC=8;}; 
MNCB[0x23]=function(){return 'SLA E';};

OPCB[0x24]=function(){var H=HL>>8;
  FC=(H>>7)&1;
  H=(H<<1)&0xFF;
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0x00FF);ICC=8;}; 
MNCB[0x24]=function(){return 'SLA H';};

OPCB[0x25]=function(){var L=HL&0xFF;
  FC=(L>>7)&1;
  L=(L<<1)&0xFF;
  FN=0;FH=0;FZ=L==0;
  HL=(HL&0xFF00)|L;ICC=8;}; 
MNCB[0x25]=function(){return 'SLA L';};

OPCB[0x26]=function(){var M=MEM[HL];
  FC=(M>>7)&1;
  M=(M<<1)&0xFF;
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;}; 
MNCB[0x26]=function(){return 'SLA (HL)';};

// SRA n
OPCB[0x2F]=function(){
  FC=RA&1;
  RA=(RA>>1)|(RA&0x80);
  FN=0;FH=0;FZ=RA==0;ICC=8;}; 
MNCB[0x2F]=function(){return 'SRA A';};

OPCB[0x28]=function(){
  FC=RB&1;
  RB=(RB>>1)|(RB&0x80);
  FN=0;FH=0;FZ=RB==0;ICC=8;}; 
MNCB[0x28]=function(){return 'SRA B';};

OPCB[0x29]=function(){
  FC=RC&1;
  RC=(RC>>1)|(RC&0x80);
  FN=0;FH=0;FZ=RC==0;ICC=8;}; 
MNCB[0x29]=function(){return 'SRA C';};

OPCB[0x2A]=function(){
  FC=RD&1;
  RD=(RD>>1)|(RD&0x80);
  FN=0;FH=0;FZ=RD==0;ICC=8;}; 
MNCB[0x2A]=function(){return 'SRA D';};

OPCB[0x2B]=function(){
  FC=RE&1;
  RE=(RE>>1)|(RE&0x80);
  FN=0;FH=0;FZ=RE==0;ICC=8;}; 
MNCB[0x2B]=function(){return 'SRA E';};

OPCB[0x2C]=function(){var H=HL>>8;
  FC=H&1;
  H=(H>>1)|(H&0x80);
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0x00FF);ICC=8;}; 
MNCB[0x2C]=function(){return 'SRA H';};

OPCB[0x2D]=function(){var L=HL&0xFF;
  FC=L&1;
  L=(L>>1)|(L&0x80);
  FN=0;FH=0;FZ=L==0;
  HL=(HL&0xFF00)|L;ICC=8;}; 
MNCB[0x2D]=function(){return 'SRA L';};

OPCB[0x2E]=function(){var M=MEM[HL];
  FC=M&1;
  M=(M>>1)|(M&0x80);
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;}; 
MNCB[0x2E]=function(){return 'SRA (HL)';};

//SRL n
OPCB[0x3F]=function(){
  FC=RA&1;
  RA=RA>>1;
  FN=0;FH=0;FZ=RA==0;ICC=8;}; 
MNCB[0x3F]=function(){return 'SRL A';};

OPCB[0x38]=function(){
  FC=RB&1;
  RB=RB>>1;
  FN=0;FH=0;FZ=RB==0;ICC=8;}; 
MNCB[0x38]=function(){return 'SRL B';};

OPCB[0x39]=function(){
  FC=RC&1;
  RC=RC>>1;
  FN=0;FH=0;FZ=RC==0;ICC=8;}; 
MNCB[0x39]=function(){return 'SRL C';};

OPCB[0x3A]=function(){
  FC=RD&1;
  RD=RD>>1;
  FN=0;FH=0;FZ=RD==0;ICC=8;}; 
MNCB[0x3A]=function(){return 'SRL D';};

OPCB[0x3B]=function(){
  FC=RE&1;
  RE=RE>>1;
  FN=0;FH=0;FZ=RE==0;ICC=8;}; 
MNCB[0x3B]=function(){return 'SRL E';};

OPCB[0x3C]=function(){var H=HL>>8;
  FC=H&1;
  H=H>>1;
  FN=0;FH=0;FZ=H==0;
  HL=(H<<8)|(HL&0x00FF);ICC=8;}; 
MNCB[0x3C]=function(){return 'SRL H';};

OPCB[0x3D]=function(){var L=HL&0xFF;
  FC=L&1;
  L=L>>1;
  FN=0;FH=0;FZ=L==0;
  HL=(HL&0xFF00)|L;ICC=8;}; 
MNCB[0x3D]=function(){return 'SRL L';};

OPCB[0x3E]=function(){var M=MEM[HL];
  FC=M&1;
  M=M>>1;
  FN=0;FH=0;FZ=M==0;
  MEM[HL]=M;ICC=16;}; 
MNCB[0x3E]=function(){return 'SRL (HL)';};

//BIT n,r
//CB 01 xxx xxx
//CB 01 bit reg
// TODO TEST THIS!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
for (var i=0;i<8;i++) {
  var o=(1<<6)|(i<<3);
  OPCB[o|7]=new Function("FZ=!(RA&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|7]=new Function("return 'BIT "+i+",A';");
  OPCB[o|0]=new Function("FZ=!(RB&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|0]=new Function("return 'BIT "+i+",B';");
  OPCB[o|1]=new Function("FZ=!(RC&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|1]=new Function("return 'BIT "+i+",C';");
  OPCB[o|2]=new Function("FZ=!(RD&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|2]=new Function("return 'BIT "+i+",D';");
  OPCB[o|3]=new Function("FZ=!(RE&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|3]=new Function("return 'BIT "+i+",E';");
  OPCB[o|4]=new Function("FZ=!(HL&"+(256<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|4]=new Function("return 'BIT "+i+",H';");
  OPCB[o|5]=new Function("FZ=!(HL&"+(1<<i)+");FN=0;FH=1;ICC=8;");
  MNCB[o|5]=new Function("return 'BIT "+i+",L';");
  OPCB[o|6]=new Function("FZ=!(MEM[HL]&"+(1<<i)+");FN=0;FH=1;ICC=16;");
  MNCB[o|6]=new Function("return 'BIT "+i+",(HL)';");
}

//SET n,r
//CB 11 xxx xxx
//CB 11 bit reg
for (var i=0;i<8;i++) {
  var o=(3<<6)|(i<<3);
  OPCB[o|7]=new Function("RA|="+(1<<i)+";ICC=8;");
  MNCB[o|7]=new Function("return 'SET "+i+",A';");
  OPCB[o|0]=new Function("RB|="+(1<<i)+";ICC=8;");
  MNCB[o|0]=new Function("return 'SET "+i+",B';");
  OPCB[o|1]=new Function("RC|="+(1<<i)+";ICC=8;");
  MNCB[o|1]=new Function("return 'SET "+i+",C';");
  OPCB[o|2]=new Function("RD|="+(1<<i)+";ICC=8;");
  MNCB[o|2]=new Function("return 'SET "+i+",D';");
  OPCB[o|3]=new Function("RE|="+(1<<i)+";ICC=8;");
  MNCB[o|3]=new Function("return 'SET "+i+",E';");
  OPCB[o|4]=new Function("HL|="+(256<<i)+";ICC=8;");
  MNCB[o|4]=new Function("return 'SET "+i+",H';");
  OPCB[o|5]=new Function("HL|="+(1<<i)+";ICC=8;");
  MNCB[o|5]=new Function("return 'SET "+i+",L';");
  OPCB[o|6]=new Function("MEM[HL]|="+(1<<i)+";ICC=16;");
  MNCB[o|6]=new Function("return 'SET "+i+",(HL)';");  
}

//RES n,r
//CB 10 xxx xxx
//CB 10 bit reg
for (var i=0;i<8;i++) {
  var o=(2<<6)|(i<<3);
  OPCB[o|7]=new Function("RA&="+((~(1<<i))&0xFF)+";ICC=8;");
  MNCB[o|7]=new Function("return 'RES "+i+",A';");
  OPCB[o|0]=new Function("RB&="+((~(1<<i))&0xFF)+";ICC=8;");
  MNCB[o|0]=new Function("return 'RES "+i+",B';");
  OPCB[o|1]=new Function("RC&="+((~(1<<i))&0xFF)+";ICC=8;");
  MNCB[o|1]=new Function("return 'RES "+i+",C';");
  OPCB[o|2]=new Function("RD&="+((~(1<<i))&0xFF)+";ICC=8;");
  MNCB[o|2]=new Function("return 'RES "+i+",D';");
  OPCB[o|3]=new Function("RE&="+((~(1<<i))&0xFF)+";ICC=8;");
  MNCB[o|3]=new Function("return 'RES "+i+",E';");
  OPCB[o|4]=new Function("HL&="+((~(256<<i))&0xFFFF)+";ICC=8;");
  MNCB[o|4]=new Function("return 'RES "+i+",H';");
  OPCB[o|5]=new Function("HL&="+((~(1<<i))&0xFFFF)+";ICC=8;");
  MNCB[o|5]=new Function("return 'RES "+i+",L';");
  OPCB[o|6]=new Function("MEM[HL]&="+((~(1<<i))&0xFF)+";ICC=16;");
  MNCB[o|6]=new Function("return 'RES "+i+",(HL)';");  
}

//JP nn
OP[0xC3]=function(){PC=(MEM[PC+1]|(MEM[PC+2]<<8))-1;ICC=12;};
MN[0xC3]=function(){return 'JP 0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//JP NZ,nn
OP[0xC2]=function(){if(!FZ)PC=(MEM[PC+1]|(MEM[PC+2]<<8))-1;else PC+=2;ICC=12;};
MN[0xC2]=function(){return 'JP NZ,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//JP Z,nn
OP[0xCA]=function(){if(FZ)PC=(MEM[PC+1]|(MEM[PC+2]<<8))-1;else PC+=2;ICC=12;};
MN[0xCA]=function(){return 'JP Z,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//JP NC,nn
OP[0xD2]=function(){if(!FC)PC=(MEM[PC+1]|(MEM[PC+2]<<8))-1;else PC+=2;ICC=12;};
MN[0xD2]=function(){return 'JP NC,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//JP C,nn
OP[0xDA]=function(){if(FC)PC=(MEM[PC+1]|(MEM[PC+2]<<8))-1;else PC+=2;ICC=12;};
MN[0xDA]=function(){return 'JP C,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//JP (HL)
OP[0xE9]=function(){PC=HL-1;ICC=4;};
MN[0xE9]=function(){return 'JP (HL)';};

//JR n
OP[0x18]=function(){PC+=sb(MEM[PC+1])+1;ICC=8;};
MN[0x18]=function(){return 'JR '+sb(MEM[PC+1])+'; 0x'+hex2(MEM[PC+1]);};

//JR NZ,n
OP[0x20]=function(){if(!FZ)PC+=sb(MEM[PC+1])+1;else PC++;ICC=8;};
MN[0x20]=function(){return 'JR NZ,'+sb(MEM[PC+1])+'; 0x'+hex2(MEM[PC+1]);};

//JR Z,n
OP[0x28]=function(){if(FZ)PC+=sb(MEM[PC+1])+1;else PC++;ICC=8;};
MN[0x28]=function(){return 'JR Z,'+sb(MEM[PC+1])+'; 0x'+hex2(MEM[PC+1]);};

//JR NC,n
OP[0x30]=function(){if(!FC)PC+=sb(MEM[PC+1])+1;else PC++;ICC=8;};
MN[0x30]=function(){return 'JR NC,'+sb(MEM[PC+1])+'; 0x'+hex2(MEM[PC+1]);};

//JR C,n
OP[0x38]=function(){if(FC)PC+=sb(MEM[PC+1])+1;else PC++;ICC=8;};
MN[0x38]=function(){return 'JR C,'+sb(MEM[PC+1])+'; 0x'+hex2(MEM[PC+1]);};

//CALL nn
//SP-1 = PCH
//SP-2 = PCL
//PUSH PC -> MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);
OP[0xCD]=function(){PC+=3;  
  MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);
  PC=((MEM[PC-1]<<8)|MEM[PC-2])-1;ICC=12;};
MN[0xCD]=function(){return 'CALL 0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//CALL NZ,nn
OP[0xC4]=function(){PC+=3;
  if(!FZ){MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=((MEM[PC-1]<<8)|MEM[PC-2])-1;}
  ICC=12;};
MN[0xC4]=function(){return 'CALL NZ,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//CALL Z,nn
OP[0xCC]=function(){PC+=3;
  if(FZ){MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=((MEM[PC-1]<<8)|MEM[PC-2])-1;}
  ICC=12;};
MN[0xCC]=function(){return 'CALL Z,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//CALL NC,nn
OP[0xD4]=function(){PC+=3;
  if(!FC){MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=((MEM[PC-1]<<8)|MEM[PC-2])-1;}
  ICC=12;};
MN[0xD4]=function(){return 'CALL NC,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//CALL C,nn
OP[0xDC]=function(){PC+=3;
  if(FC){MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=((MEM[PC-1]<<8)|MEM[PC-2])-1;}
  ICC=12;};
MN[0xDC]=function(){return 'CALL C,0x'+hex(MEM[PC+1]|(MEM[PC+2]<<8));};

//RST
//PUSH PC -> MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);
OP[0xC7]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=  -1;ICC=32;};
MN[0xC7]=function(){return 'RST 0x00';};

OP[0xCF]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x07;ICC=32;};
MN[0xCF]=function(){return 'RST 0x08';};

OP[0xD7]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x0F;ICC=32;};
MN[0xD7]=function(){return 'RST 0x10';};

OP[0xDF]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x17;ICC=32;};
MN[0xDF]=function(){return 'RST 0x18';};

OP[0xE7]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x1F;ICC=32;};
MN[0xE7]=function(){return 'RST 0x20';};

OP[0xEF]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x27;ICC=32;};
MN[0xEF]=function(){return 'RST 0x28';};

OP[0xF7]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x2F;ICC=32;};
MN[0xF7]=function(){return 'RST 0x30';};

OP[0xFF]=function(){PC++;MEMW(--SP,PC>>8);MEMW(--SP,PC&0xFF);PC=0x37;ICC=32;};
MN[0xFF]=function(){return 'RST 0x38';};

//RET
//POP PC ->  var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;
OP[0xC9]=function(){PC=(MEM[SP++]|(MEM[SP++]<<8))-1;ICC=8;};
MN[0xC9]=function(){return 'RET';};

//RET NZ
OP[0xC0]=function(){if(!FZ){var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;}ICC=8;};
MN[0xC0]=function(){return 'RET NZ';};

//RET Z
OP[0xC8]=function(){if(FZ){var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;}ICC=8;};
MN[0xC8]=function(){return 'RET Z';};

//RET NC
OP[0xD0]=function(){if(!FC){var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;}ICC=8;};
MN[0xD0]=function(){return 'RET NC';};

//RET C
OP[0xD8]=function(){if(FC){var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;}ICC=8;};
MN[0xD8]=function(){return 'RET C';};

//RETI
OP[0xD9]=function(){var L=MEM[SP++];PC=((MEM[SP++]<<8)|L)-1;IME=true;ICC=8;};
MN[0xD9]=function(){return 'RETI';};

//------------------------------------------------------------------------------

MN[0x3E]=function(){return 'LD A,0x'+hex2(MEM[PC+1]);}; // ???
MN[0x06]=function(){return 'LD B,0x'+hex2(MEM[PC+1]);};
MN[0x0E]=function(){return 'LD C,0x'+hex2(MEM[PC+1]);};
MN[0x16]=function(){return 'LD D,0x'+hex2(MEM[PC+1]);};
MN[0x1E]=function(){return 'LD E,0x'+hex2(MEM[PC+1]);};
MN[0x26]=function(){return 'LD H,0x'+hex2(MEM[PC+1]);};
MN[0x2E]=function(){return 'LD L,0x'+hex2(MEM[PC+1]);};
MN[0x7F]=function(){return 'LD A,A';};
MN[0x78]=function(){return 'LD A,B';};
MN[0x79]=function(){return 'LD A,C';};
MN[0x7A]=function(){return 'LD A,D';};
MN[0x7B]=function(){return 'LD A,E';};
MN[0x7C]=function(){return 'LD A,H';};
MN[0x7D]=function(){return 'LD A,L';};
MN[0x7E]=function(){return 'LD A,(HL)';};
MN[0x47]=function(){return 'LD B,A';};
MN[0x40]=function(){return 'LD B,B';};
MN[0x41]=function(){return 'LD B,C';};
MN[0x42]=function(){return 'LD B,D';};
MN[0x43]=function(){return 'LD B,E';};
MN[0x44]=function(){return 'LD B,H';};
MN[0x45]=function(){return 'LD B,L';};
MN[0x46]=function(){return 'LD B,(HL)';};
MN[0x4F]=function(){return 'LD C,A';};
MN[0x48]=function(){return 'LD C,B';};
MN[0x49]=function(){return 'LD C,C';};
MN[0x4A]=function(){return 'LD C,D';};
MN[0x4B]=function(){return 'LD C,E';};
MN[0x4C]=function(){return 'LD C,H';};
MN[0x4D]=function(){return 'LD C,L';};
MN[0x4E]=function(){return 'LD C,(HL)';};
MN[0x57]=function(){return 'LD D,A';};
MN[0x50]=function(){return 'LD D,B';};
MN[0x51]=function(){return 'LD D,C';};
MN[0x52]=function(){return 'LD D,D';};
MN[0x53]=function(){return 'LD D,E';};
MN[0x54]=function(){return 'LD D,H';};
MN[0x55]=function(){return 'LD D,L';};
MN[0x56]=function(){return 'LD D,(HL)';};
MN[0x5F]=function(){return 'LD E,A';};
MN[0x58]=function(){return 'LD E,B';};
MN[0x59]=function(){return 'LD E,C';};
MN[0x5A]=function(){return 'LD E,D';};
MN[0x5B]=function(){return 'LD E,E';};
MN[0x5C]=function(){return 'LD E,H';};
MN[0x5D]=function(){return 'LD E,L';};
MN[0x5E]=function(){return 'LD E,(HL)';};
MN[0x67]=function(){return 'LD H,A';};
MN[0x60]=function(){return 'LD H,B';};
MN[0x61]=function(){return 'LD H,C';};
MN[0x62]=function(){return 'LD H,D';};
MN[0x63]=function(){return 'LD H,E';};
MN[0x64]=function(){return 'LD H,H';};
MN[0x65]=function(){return 'LD H,L';};
MN[0x66]=function(){return 'LD H,(HL)';};
MN[0x6F]=function(){return 'LD L,A';};
MN[0x68]=function(){return 'LD L,B';};
MN[0x69]=function(){return 'LD L,C';};
MN[0x6A]=function(){return 'LD L,D';};
MN[0x6B]=function(){return 'LD L,E';};
MN[0x6C]=function(){return 'LD L,H';};
MN[0x6D]=function(){return 'LD L,L';};
MN[0x6E]=function(){return 'LD L,(HL)';};
MN[0x70]=function(){return 'LD (HL),B';};
MN[0x71]=function(){return 'LD (HL),C';};
MN[0x72]=function(){return 'LD (HL),D';};
MN[0x73]=function(){return 'LD (HL),E';};
MN[0x74]=function(){return 'LD (HL),H';};
MN[0x75]=function(){return 'LD (HL),L';};
MN[0x36]=function(){return 'LD (HL),0x'+hex2(MEM[PC+1]);};
MN[0x0A]=function(){return 'LD A,(BC)';};
MN[0x1A]=function(){return 'LD A,(DE)';};
MN[0x7E]=function(){return 'LD A,(HL)';};
MN[0xFA]=function(){return 'LD A,(0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1])+')';};
MN[0x02]=function(){return 'LD (BC),A';};
MN[0x12]=function(){return 'LD (DE),A';};
MN[0x77]=function(){return 'LD (HL),A';};
MN[0xEA]=function(){return 'LD (0x'+hex((MEM[PC+2]<<8)|MEM[PC+1],4)+'),A';}; 
MN[0x08]=function(){return 'LD(0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1])+'),SP';}; 
MN[0xF8]=function(){return 'LD HL,SP+0x'+hex2(MEM[PC+1]);};
MN[0xF9]=function(){return 'LD SP,HL';};
MN[0x31]=function(){return 'LD SP,0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1]);};
MN[0x21]=function(){return 'LD HL,0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1]);};
MN[0x11]=function(){return 'LD DE,0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1]);};
MN[0x01]=function(){return 'LD BC,0x'+hex4((MEM[PC+2]<<8)+MEM[PC+1]);};
MN[0xF0]=function(){return 'LD A,(0xFF00+0x'+hex2(MEM[PC+1])+')';};
MN[0xE0]=function(){return 'LD (0xFF00+0x'+hex2(MEM[PC+1])+'),A';};
MN[0xF2]=function(){return 'LD A,(0xFF00+C)';};
MN[0xE2]=function(){return 'LD (0xFF00+C),A';};
MN[0x3A]=function(){return 'LDD A,(HL)';};
MN[0x32]=function(){return 'LDD (HL),A';};
MN[0x2A]=function(){return 'LDI A,(HL)';};
MN[0x22]=function(){return 'LDI (HL),A';};
MN[0xF5]=function(){return 'PUSH AF';};
MN[0xC5]=function(){return 'PUSH BC';};
MN[0xD5]=function(){return 'PUSH DE';};
MN[0xE5]=function(){return 'PUSH HL';};

function init_cpu() {
  RA=0x01; // 0x01->GB/SGB; 0xFF->GBP; 0x11->GBC
  FZ=0x01; // F=0xB0->Z1 N0 H1 C1
  FN=0x00;
  FH=0x01;
  FC=0x01;
  RB=0x00;
  RC=0x13;
  RD=0x00;
  RE=0xD8;
  PC=0x0100;
  SP=0xFFFE;
  HL=0x014D;
  ICC=0;
  CCC=0;
}  
init_cpu();


