/* 
 * jsgb.graphics.js v0.02 - Graphic functions for JSGB, a GameBoy Emulator
 * Copyright (C) 2009 Pedro Ladaria <Sonic1980 at Gmail dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The full license is available at http://www.gnu.org/licenses/gpl.html
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 */

var SLC = 0;    // ScanLine Counter
var TOBJS = []; // tile canvas objects
var TCTXS = []; // tile canvas contexts
var BGOBJ;      // background canvas object
var BGCTX;      // background canvascontext
var SPOBJ;      // sprites Object
var SPCTX;      // sprites Context
var LCDOBJ;     // LCD Object
var LCDCTX;     // LCD Context

var UPDT  = false;// Update tiles?
var UPDTL = [];   // Update Tile List. Set to True when VRAM tile data changes
                  // so, on each frame we only need to redraw modified tiles.
                  // This is set on MEMW function (jsgb.memory.js)
var UPDB  = false // Update Background?
var UPDBL = [];   // Update Background List                      
var UPDBF = [];   // Force update background array;
var UPDS  = true; // Update Sprites


var LINES  = []; // conversion table: 2 bytes -> 8 pixels
var COLORS = [[0x00,0x00,0x00,0x00],[0xD0,0xD0,0xD0,0xFF],
              [0xA0,0xA0,0xA0,0xFF],[0x50,0x50,0x50,0xFF]];

// tests for ctx.putImageData availability
var can_putImageData = true;
function putImageData_test() {
  can_putImageData = true;
  try { LCDCTX.putImageData({width:1,height:1,data:[0,0,0,0]},0,0); }  
  catch(shit) { can_putImageData = false; }    
}

// ctx.putImageData alternative for browsers that doesn't supports it
function putImageDataByRects(ctx, w, h, data) {
  var d=0;
  for (var y=0; y<h; y++) for (var x=0; x<w; x++) {
    var a=data[d+3]/255;
    if (a>0) {
      ctx.fillStyle = 'rgba('+data[d]+','+data[d+1]+','+data[d+2]+','+a+')';
      ctx.fillRect(x,y,1,1);
    }  
    d+=4;
  }
}

(function init_lines_table() {
  for (var i=0; i<0x10000; i++) {
    LINES[i]=[]; // line (8 pixels)
    LINES[i]=LINES[i].concat(
      COLORS[((i>>7)&1)|((i>>14)&2)], // bits 7 and 15
      COLORS[((i>>6)&1)|((i>>13)&2)], // bits 6 and 14
      COLORS[((i>>5)&1)|((i>>12)&2)], // bits 5 and 13
      COLORS[((i>>4)&1)|((i>>11)&2)], // bits 4 and 12
      COLORS[((i>>3)&1)|((i>>10)&2)], // bits 3 and 11
      COLORS[((i>>2)&1)|((i>> 9)&2)], // bits 2 and 10
      COLORS[((i>>1)&1)|((i>> 8)&2)], // bits 1 and 9
      COLORS[((i>>0)&1)|((i>> 7)&2)]  // bits 0 and 8
    );  
  }
})();

function draw_tiles() {
  var b=(MEM[0xFF40]>>4)&1; // get LCDC bit 4
  var tda=0; // tile data addr
  var pixels; // tile pixels
  var d=0; // data index
  var l=0; // line (2 bytes)
  for (var i=0;i<384;i++) if (UPDTL[i]) { // loop tiles and redraw if needed
    pixels=[];
    tda=0x8000+i*16;
    for (var j=0;j<8;j++) { // loop 8 lines    
      l =MEM[tda++]<<8;
      l|=MEM[tda++];
      pixels=pixels.concat(LINES[l]);
    }
    // clear
    TCTXS[i].clearRect(0,0,8,8);
    if (can_putImageData) { // Firefox 3+, etc
      TCTXS[i].putImageData({width:8,height:8,data:pixels},0,0);
    }
    else { // Chrome, etc
      putImageDataByRects(TCTXS[i],8,8,pixels);
    }
    UPDTL[i]=false;
  }
  UPDT=false; 
}

function draw_background() {
  // TODO this could be improved looping like "for (var i in UPDBL)"
  // and at the end make UPDBL=[]; on MEMW set values to true when needed
  // and do NOT initialize array on init_graphics();
  // the same for draw_tiles();
  // TODO when LCDC bit3/4 changes, all the background must be repainted
/*
Bit 4 - BG & Window Tile Data Select
         0: $8800-$97FF
         1: $8000-$8FFF

Bit 3 - BG Tile Map Display Select
         0: $9800-$9BFF
         1: $9C00-$9FFF
*/
  var x=0;
  var y=0;
  var BTM=(MEM[0xFF40]&8)?0x9C00:0x9800; // (LCDC&8)?BTM1:BTM0 - bg tile map
  if (MEM[0xFF40]&16) {  // background tile data
    for (var i=0;i<2048;i++) if (UPDBL[i]) { 
      x=(i%32)*8;
      y=(i>>5)*8;
      BGCTX.fillRect(x,y,8,8);
      BGCTX.drawImage(TOBJS[MEM[BTM+i]],x,y);
      UPDBL[i]=false;
    }  
  }
  else { 
    for (var i=0;i<2048;i++) if (UPDBL[i]) { 
      x=(i%32)*8;
      y=(i>>5)*8;
      BGCTX.fillRect(x,y,8,8);
      BGCTX.drawImage(TOBJS[256+sb(MEM[BTM+i])],x,y);
      UPDBL[i]=false;
    }  
  }
  UPDB = false;  
}

/*
Byte0 Y position on the screen
Byte1 X position on the screen
Byte2 Pattern number 0-255 (Unlike some tile numbers, sprite pattern numbers
      are unsigned. LSB is ignored (treated as 0) in 8x16 mode.)
Byte3 Flags:
  Bit7 Priority: If this bit is set to 0, sprite is displayed on top of
       background & window. If this bit is set to 1, then sprite will be hidden
       behind colors 1, 2, and 3 of the background & window. (Sprite only
       prevails over color 0 of BG & win.)
  Bit6 Y flip: Sprite pattern is flipped vertically if this bit is set to 1.
  Bit5 X flip: Sprite pattern is flipped horizontally if this bit is set to 1.
  Bit4 Palette number: Sprite colors are taken from OBJ1PAL if this bit is set
       to 1 and from OBJ0PAL otherwise.
*/
function draw_sprites() {
  // TODO A LOT!!
  SPCTX.clearRect(0,0,160,144);  
  for (var i=0;i<160;i++) {
    var y=MEM[0xFE00+i++];
    var x=MEM[0xFE00+i++];
    var p=MEM[0xFE00+i++];
    var f=MEM[0xFE00+i];
    if ((y>0)&(x>0)) SPCTX.drawImage(TOBJS[p],x-8,y-16);
  }
}

var tSCY=0; // save SCY for each frame because it should change on VBlank only
var ENDFRAME = false;

function draw_scanline() {
  // TODO check LCDC values before painting
  var LY=MEM[0xFF44];
  if (LY==0) {
    tSCY=MEM[0xFF42];
    if (UPDT) draw_tiles();          // TODO this is not correct.
    if (UPDS) draw_sprites();        // data can change outside here
    if (UPDB) draw_background();     // but here it runs faster...
  } 
  // TODO Y sould be calculated on MEMW to gain a little speed
  // TODO paint X overflows
  var Y=(MEM[0xFF40]&8)?256:0; // get map start addr from LCDC
  LCDCTX.drawImage(BGOBJ, MEM[0xFF43],Y+((tSCY+LY)%256), 160,1, 0,LY, 160,1);
  LCDCTX.drawImage(SPOBJ, 0,LY, 160,1, 0,LY, 160,1);  
}

function init_graphics() {
  // create canvases
  var s='';
  var id=0;
  var n=0;
  var t='';
  for (var j=0;j<24;j++) {
    for (var i=0;i<16;i++) {
      n=(j*16)+i;
      t='#'+n+' (0x'+hex2(n)+') @ 0x'+hex4(0x8000+(n*16));
      s+='<canvas width="8" height="8" id="TILE_'+id+'" title="'+t+'"></canvas>';
      id++;
    }  
    s+=br;
  }
  $('TILES').innerHTML=s;
  // store canvas objects into arrays and get 2D contexts
  for (var i=0;i<384;i++) {
    TOBJS[i]=$('TILE_'+i);
    TCTXS[i]=TOBJS[i].getContext('2d');
  }  
  // init LCD Screen variables
  LCDOBJ=$('LCD');
  LCDCTX=LCDOBJ.getContext('2d');
  LCDCTX.fillStyle='rgba(255,255,255,1)';
  LCDCTX.fillRect(0,0,160,144);
  // init background variables
  BGOBJ=$('BG_CANVAS');
  BGCTX=BGOBJ.getContext('2d');
  BGCTX.fillStyle='rgba(255,255,255,1)';
  BGCTX.fillRect(0,0,256,512);
  // init sprite variables
  SPOBJ=$('SP_CANVAS');
  SPCTX=SPOBJ.getContext('2d');
  // test putImageData function
  putImageData_test();
  // update tiles info
  UPDT = false;
  for (var i=0; i<384; i++) UPDTL[i]=false;   
  // update bg info
  UPDB = false;
  for (var i=0; i<2048; i++) {
    UPDBL[i]=false;
    UPDBF[i]=true;
  }  
}

function MODE0() { // H-Blank 
  if ((MEM[0xFF41]&3)!=0) {
    MEM[0xFF41]&=0xFC; // set STAT bits 1-0 to 0
    if(MEM[0xFF41]&8) MEM[0xFF0F]|=2; // if STAT bit 3 -> set IF bit1
  }  
}

function MODE2() { // OAM in use
  if ((MEM[0xFF41]&3)!=2) {
    MEM[0xFF41]=(MEM[0xFF41]&0xFC)|2;// set STAT bits 1-0 to 2
    if(MEM[0xFF41]&32) MEM[0xFF0F]|=2;// if STAT bit 5 -> set IF bit1
    // TODO here sould sprites be updated?
  }  
}

function MODE3() { // OAM+VRAM busy
  if ((MEM[0xFF41]&3)!=3) {
    MEM[0xFF41]|=3; // set STAT bits 1-0 to 3
    if (MEM[0xFF40]&128) draw_scanline(); // look at 0xFF40 for LCD state
  }  
}

function MODE1() { // V-Blank  
  MEM[0xFF41]=(MEM[0xFF41]&0xFC)|1;
  MEM[0xFF0F]|=1; // set IF flag 0 
}

function LY_LYC() { // LY - LYC Compare
  if (MEM[0xFF44]==MEM[0xFF45]) { // If LY==LCY
    MEM[0xFF41]|=0x04; // set STAT bit 2         
    if (MEM[0xFF41]&64) MEM[0xFF0F]|=2; // if STAT bit 6 -> set IF bit1
  }      
  else MEM[0xFF41]&=0xFB; // reset STAT bit 2 (LY!=LYC)
}

function LCD_CONTROL() {
  SLC+=ICC; // ScanLineCounter += InstructionCyclesCount
  if(SLC>455){ // when ScanLineCounter overflows we are in a new scanline    
    SLC-=456;
    MEM[0xFF44]++; // inc LY (current scanline)
    if(MEM[0xFF44]>153)MEM[0xFF44]-=154;
    if (MEM[0xFF44]==144) MODE1(); // Mode1: 4560 cycles
    else if(MEM[0xFF44]==0) ENDFRAME=true; 
    LY_LYC();
  }
  if (MEM[0xFF44]<144) { // if not in V-Blank
    if (SLC<205) MODE0(); // Mode0: 204 cycles
    else if(SLC<285) MODE2(); // Mode2: 80 cycles
    else MODE3(); // Mode3: 172 cycles
  }
}

