/* 
 * jsgb.input.js v0.02 - Input module for JSGB, a GameBoy Emulator
 * Copyright (C) 2009 Pedro Ladaria <Sonic1980 at Gmail dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The full license is available at http://www.gnu.org/licenses/gpl.html
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 */
 
/*
  Bit 7 - Gameboy ID (0=super gameboy, 1=gameboy)
  Bit 6 - Gameboy ID (0=super gameboy, 1=gameboy)
  Bit 5 - P15 out port ---------+        (0=select)
  Bit 4 - P14 out port --+      |
  Bit 3 - P13 in port   down   start     (0=pressed)
  Bit 2 - P12 in port   up     select
  Bit 1 - P11 in port   left   button-B
  Bit 0 - P10 in port   right  button-A
*/

var P14; // up down left right
var P15; // start select a b

function setP1(v) {
  switch ((v>>4)&3) {
    case 0: MEM[0xFF00]=P14&P15; break; // TODO not sure on this
    case 1: MEM[0xFF00]=P15; break;
    case 2: MEM[0xFF00]=P14; break;
    case 3: MEM[0xFF00]=0xFF; break; // TODO not sure on this
  }
}
    
function gbkeydown(e) {  
  switch (e.which) {
    case 40: P14&=0xF7; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 38: P14&=0xFB; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 37: P14&=0xFD; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 39: P14&=0xFE; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 65: P15&=0xF7; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 83: P15&=0xFB; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 88: P15&=0xFD; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 90: P15&=0xFE; MEM[0xFF0F]|=16; e.preventDefault(); break;
  }
}

function gbkeyup(e) {
  switch (e.which) {
    case 40: P14|=8; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 38: P14|=4; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 37: P14|=2; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 39: P14|=1; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 65: P15|=8; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 83: P15|=4; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 88: P15|=2; MEM[0xFF0F]|=16; e.preventDefault(); break;
    case 90: P15|=1; MEM[0xFF0F]|=16; e.preventDefault(); break;
  }
}

function init_input() {
  document.onkeydown = gbkeydown;
  document.onkeyup = gbkeyup;
  P14=0xEF;
  P15=0xDF;
}  

