/* 
 * jsgb.interrupts.js v0.01 - Interrupt handling for JSGB, a GameBoy Emulator
 * Copyright (C) 2009 Pedro Ladaria <Sonic1980 at Gmail dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The full license is available at http://www.gnu.org/licenses/gpl.html
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 */
 
/*
0040 0 Vertical Blank Interrupt Start Address
0048 1 LCDC Status Interrupt Start Address
0050 2 Timer Overflow Interrupt Start Address
0058 3 Serial Transfer Completion Interrupt Start Address
0060 4 High-to-Low of P10-P13 Interrupt Start Address
*/

function INT_VBLANK() { // IF/IE bit 0
  IME=false;
  MEM[0xFF0F]&=0xFE;  // reset IF bit 0
  MEMW(--SP,PC>>8);   // store PCH
  MEMW(--SP,PC&0xFF); // store PCL
  PC =0x0040;         // jump!
  ICC+=32;
}

function INT_STAT() { // IF/IE bit 1
  IME=false;
  MEM[0xFF0F]&=0xFD;  // reset IF bit 1
  MEMW(--SP,PC>>8);   // store PCH
  MEMW(--SP,PC&0xFF); // store PCL
  PC =0x0048;         // jump!
  ICC+=32;
}

function INT_TIMER() { // IF/IE bit 2
  IME=false;
  MEM[0xFF0F]&=0xFB;  // reset IF bit 2
  MEMW(--SP,PC>>8);   // store PCH
  MEMW(--SP,PC&0xFF); // store PCL
  PC =0x0050;         // jump!
  ICC+=32;
}

function INT_SERIAL() { // IF/IE bit 3
  IME=false;
  MEM[0xFF0F]&=0xF7;  // reset IF bit 3
  MEMW(--SP,PC>>8);   // store PCH
  MEMW(--SP,PC&0xFF); // store PCL
  PC =0x0058;         // jump!
  ICC+=32;
}

function INT_BUTTONS() { // IF/IE bit 4
  IME=false;
  MEM[0xFF0F]&=0xEF;  // reset IF bit 4
  MEMW(--SP,PC>>8);   // store PCH
  MEMW(--SP,PC&0xFF); // store PCL
  PC =0x0060;         // jump!
  ICC+=32;
}

function INT_NONE(){}

var INT = [];
function init_interrupts() {
  for (i=0;i<32;i++) {
    if (i&1) INT[i]=INT_VBLANK; else    
    if (i&2) INT[i]=INT_STAT; else    
    if (i&4) INT[i]=INT_TIMER; else    
    if (i&8) INT[i]=INT_SERIAL; else    
    if (i&16)INT[i]=INT_BUTTONS; else
    INT[i]=INT_NONE;   
  }
}
