/* 
 * jsgb.debugger.js v0.02 - Memory module for JSGB, a GameBoy Emulator
 * Copyright (C) 2009 Pedro Ladaria <Sonic1980 at Gmail dot com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The full license is available at http://www.gnu.org/licenses/gpl.html
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 */
 
var MEM = new Array(0x10000);

/*
// special register addresses
var _SC_   = 0xFF02;
var _DIV_  = 0xFF04;
var _TIMA_ = 0xFF05;
var _TMA_  = 0xFF06;
var _TAC_  = 0xFF07;
var _IF_   = 0xFF0F;
var _LCDC_ = 0xFF40;
var _STAT_ = 0xFF41;
var _SCY_  = 0xFF42;
var _SCX_  = 0xFF43;
var _LY_   = 0xFF44;
var _LYC_  = 0xFF45;
var _DMA_  = 0xFF46;
var _BGP_  = 0xFF47;
var _OBP0_ = 0xFF48;
var _OBP1_ = 0xFF49;
var _IE_   = 0xFFFF;

// start addresses
var _ROM0_ = 0x0000;
var _ROM1_ = 0x4000; 
var _VRAM_ = 0x8000; // video RAM
var _BTD0_ = 0x8000; // backgroun tile data 0
var _BTD1_ = 0x8800; // backgroun tile data 1
var _BTM0_ = 0x9800; // background tile map 0
var _BTM1_ = 0x9C00; // background tile map 1
var _RAM1_ = 0xA000; // switchable RAM
var _RAM0_ = 0xC000; // internal RAM
var _ECHO_ = 0xE000; // echo of internal RAM
var _OAM_  = 0xFE00; // object attribute
*/

function where_mem(a) { // TODO rewrite this
  if (a<0x4000) return 'ROM0'; else
  if (a<0x8000) return 'ROM1'; else
  if (a<0xA000) return 'VRAM'; else
  if (a<0xC000) return 'RAM1'; else
  if (a<0xE000) return 'RAM0'; else
  if (a<0xFE00) return 'ECHO'; else
  if (a<0xFEA0) return 'OAM&nbsp;'; else
  if (a<0xFF00) return 'I/O&nbsp;'; else
  if (a<0xFF4C) return 'I/O&nbsp;'; else
  if (a<0xFF80) return 'I/O&nbsp;'; else
  if (a<0xFFFF) return 'HRAM'; else
  if (a=0xFFFF) return 'IE&nbsp;&nbsp;'; else
  return '&nbsp;&nbsp;&nbsp;&nbsp;';
}

function init_mem() {
  for (var i=0;i<=0xFFFF;i++) MEM[i]=0;  
  MEM[0xFF00]=0xFF; // P1
  MEM[0xFF04]=0xAF; // DIV
  MEM[0xFF05]=0x00; // TIMA
  MEM[0xFF06]=0x00; // TMA
  MEM[0xFF07]=0x00; // TAC
  MEM[0xFF0F]=0x00; // IF 
  MEM[0xFF10]=0x80; // NR10
  MEM[0xFF11]=0xBF; // NR11
  MEM[0xFF12]=0xF3; // NR12
  MEM[0xFF14]=0xBF; // NR14
  MEM[0xFF16]=0x3F; // NR21
  MEM[0xFF17]=0x00; // NR22
  MEM[0xFF19]=0xBF; // NR24
  MEM[0xFF1A]=0x7F; // NR30
  MEM[0xFF1B]=0xFF; // NR31
  MEM[0xFF1C]=0x9F; // NR32
  MEM[0xFF1E]=0xBF; // NR33
  MEM[0xFF20]=0xFF; // NR41
  MEM[0xFF21]=0x00; // NR42
  MEM[0xFF22]=0x00; // NR43
  MEM[0xFF23]=0xBF; // NR30
  MEM[0xFF24]=0x77; // NR50
  MEM[0xFF25]=0xF3; // NR51
  MEM[0xFF26]=0xF1; // NR52 0xF1->GB; 0xF0->SGB
  MEM[0xFF40]=0x91; // LCDC
  MEM[0xFF42]=0x00; // SCY
  MEM[0xFF43]=0x00; // SCX
  MEM[0xFF44]=0x00; // LY
  MEM[0xFF45]=0x00; // LYC
  MEM[0xFF47]=0xFC; // BGP
  MEM[0xFF48]=0xFF; // OBP0
  MEM[0xFF49]=0xFF; // OBP1
  MEM[0xFF4A]=0x00; // WY
  MEM[0xFF4B]=0x00; // WX
  MEM[0xFFFF]=0x00; // IE
}  

function MEMW(a,v) {
  // Special registers+HRAM
  if (a>=0xFF00) switch(a){
    // FF00 P1 Joypad
    case 0xFF00: 
    //if(v==3)MEM[a]=0xF0; else // Fx->GB/GBP; 3x->SGB
      setP1(v);
      break;
    // FF02 SC
    case 0xFF02:
      MEM[0xFF02]=0;
//      pause();
//      alert('serial transfer');
      break;
    // FF04 DIV  
    case 0xFF04:
      MEM[a]=0; // Writing any value sets it to 0.
      break;
    // FF40 LCDC  
    case 0xFF40:
      if((MEM[0xFF40]&24)!=(v&24)) { // If bits 3 or 4 gets changed, redraw BG
        UPDBL=UPDBF;
        UPDB=true;
      }
      MEM[a]=v;      
      if(!(v&128)) { // Bit 7: LCD Control Operation (0:Off/1:On)
        MEM[0xFF44]=0; // LY = 0;
      }
      break;
    // FF44 LY  
    case 0xFF44:
      MEM[a]=0; // Writing any value sets it to 0.
      break;
    // FF46 DMA TRANSFER  
    case 0xFF46:
      a=v<<8; // start address of DMA
      for (var i=0;i<160;i++) {
        MEM[0xFE00+i]=MEM[a+i];
        UPDS=true;
      }
     // pause();
      break;
    // FF05 TIMA 
    case 0xFF05:
      MEM[0xFF05]=v;
      break;  
    // FF06 TMA 
    case 0xFF06:
      MEM[0xFF06]=v;
      break;  
    // FF07 TIMER CONTROL  
    case 0xFF07: 
      MEM[a]=v;
      switch(v&3) {
        case 0:TMC=1024;break; //4.096 KHz
        case 1:TMC=16;  break; //262.144 Khz
        case 2:TMC=64;  break; //65.536 KHz
        case 3:TMC=256; break; //16.384 KHz
      }          
      break;    
    // THE OTHERS
    default: 
      MEM[a]=v;
      break;
  }
  // writing to ROM?
  else if (a<0x8000) {
    //pause();
    //alert('Warning!\n\nWriting to ROM:\nPC: 0x'+hex(PC)+'\nMEM[0x'+hex(a)+']=0x'+hex(v));
  }
  // only make changes if new value is different
  else if (MEM[a]!=v) {
    // TODO here are too many compares, remove half of them
    // Mark tile data changes
    if ((a>0x8000)&&(a<0x9800)) {
      UPDT=true;
      UPDTL[(a-0x8000)>>4]=true;
      MEM[a]=v;
    }
    // Mark tile map changes
    else if ((a>=0x9800)&&(a<0xA000)) {
      UPDB=true;
      UPDBL[a-0x9800]=true;
      MEM[a]=v;
    }  
    // Mark OAM changes
    else if ((a>=0xFE00)&&(a<0xFEA0)) {
      UPDS=true;
      MEM[a]=v;
    }
    // Writing to RAM0-RAM+1E00 also writes ECHO
    else if ((a>=0xC000)&&(a<0xDE00)) {
      MEM[a]=v;
      MEM[a+0x2000]=v;
    }
    // Writing to ECHO also writes RAM0
    else if ((a>=0xE000)&&(a<0xFE00)) {
      MEM[a]=v;
      MEM[a-0x2000]=v;
    }
    // Everything else...
    else MEM[a]=v;
  }  
}

